#pragma once
#include <vector>
#include <list>

#include "exception.h"
#include "vec2n.h"
#include "bmp.h"


template <typename T>
class MatrixData
{
    typedef T type_t;
    Vec2n m_size;
    int m_countRefs;
    std::vector<type_t> m_vecData;
public:
    MatrixData ()
    {
        m_countRefs = 1;
        m_size.x() = 0;
        m_size.y() = 0;
    }
    MatrixData (MatrixData<T>& md)
    {
        m_countRefs = 1;
        SetSize (md.GetSize ());
        for (int y = 0; y < m_size.y(); y++)
            for (int x = 0; x < m_size.x(); x++)
            {
                Vec2n v (x, y);
                type_t value = md.Get (v);
                Set (v, value);
            }
    }
    int GetOffset (Vec2n pos)
    {
        if (! (pos.x() < m_size.x()))
            throw new ExceptionInvalidParam ("Param pos is invalid");
        if (! (pos.y() < m_size.y()))
            throw new ExceptionInvalidParam ("Param pos is invalid");
        return pos.x() + pos.y()*m_size.x();
    }
    void SetSize (Vec2n& size)
    {
        m_size = size;
        m_vecData.resize (size.x()*size.y());
    }
    Vec2n GetSize ()
    {
        return m_size;
    }
    type_t Get (Vec2n& pos)
    {
        return m_vecData.at(GetOffset(pos));
    }
    void Set (Vec2n& pos, type_t value)
    {
        m_vecData.at(GetOffset(pos)) = value;
    }
    int Ref ()
    {
        return ++m_countRefs;
    }
    int Unref ()
    {
        int refs = --m_countRefs;
        if (refs==0)
            delete this;
        return refs;
    }
};


template <typename T>
class Matrix
{
public:
    typedef T type_t;
private:
    friend Matrix<T>;
    Vec2n m_org, m_size;
    Vec2n m_dirX, m_dirY;
    MatrixData<T>* m_pdata;
    
public:
    Matrix ()
    {
        m_pdata = NULL;
        m_size = Vec2n (0, 0);
        m_org = Vec2n (0, 0);
        m_dirX = Vec2n (1, 0);
        m_dirY = Vec2n (0, 1);
    }
    Matrix<T> (Matrix<T>& mat)
    {
        m_pdata = NULL;
        Clear ();
        *this = mat;
    }
    ~Matrix<T> ()
    {
        if (m_pdata)
            m_pdata->Unref ();
    }
    void Clear ()
    {
        if (m_pdata)
        {
            m_pdata->Unref();
            m_pdata = NULL;
        }
        m_size = Vec2n (0, 0);
        m_org = Vec2n (0, 0);
        m_dirX = Vec2n (1, 0);
        m_dirY = Vec2n (0, 1);
    }
    void Recreate (Vec2n& size)
    {
        m_org = Vec2n (0,0);
        m_size = size;

        if (m_pdata)
            m_pdata->Unref();
        m_pdata = new MatrixData<type_t> ();
        m_pdata->SetSize (size);
    }
    void Recreate (int x, int y)
    {
        Recreate (Vec2n (x, y));
    }
    Vec2n GetSize ()
    {
        return m_size;
    }
private:
    Vec2n GetPosMatrixData (Vec2n& in)
    {
        if (in.x() >= m_size.x() && in.y() >= m_size.y())
            throw new ExceptionInvalidParam ("Param in is invalid");
        
        Vec2n org = m_org;
        Vec2n offset = m_dirX*in.x() + m_dirY*in.y();
        return org + offset;
    }
public:
    type_t GetValue (Vec2n& pos)
    {
        if (! m_pdata)
            throw new ExceptionFatalError ("Matrix is not init");
        
        return m_pdata->Get (GetPosMatrixData(pos)); 
    }
    type_t GetValue (int x, int y)
    {
        return GetValue (Vec2n (x, y));
    }

    Matrix<type_t> GetVerticalFlippedMatrix ()
    {
        Matrix<type_t> mat = *this;
        mat.m_org.y() = mat.m_org.y() + (mat.m_size.y()-1)*(mat.m_dirX.y()+mat.m_dirY.y());
        mat.m_dirX.y() = -mat.m_dirX.y();
        mat.m_dirY.y() = -mat.m_dirY.y();
        
        return mat;
    }
    Matrix<type_t> GetRotated ()
    {
        Matrix<type_t> mat = * this;
        Vec2n v2offset (m_size.x() ? m_size.x()-1 : 0, m_size.y() ? m_size.y()-1 : 0);
        mat.m_org += Vec2n (m_dirX.x()*v2offset.x(), m_dirX.y()*v2offset.x());
        mat.m_size = Vec2n (this->m_size.y(), this->m_size.x());
        mat.m_dirX = Vec2n (+m_dirY.x(), +m_dirY.y());
        mat.m_dirY = Vec2n (-m_dirX.x(), -m_dirX.y());
        return mat;
    }
    void SetValue (Vec2n& pos, type_t value)
    {
        if (! m_pdata)
            throw new ExceptionFatalError ("Matrix is not init");
        m_pdata->Set (GetPosMatrixData(pos), value);
    }
    void SetValue (int x, int y, type_t value)
    {
        SetValue (Vec2n (x, y), value);
    }
    void PrintToStdout ()
    {
        for (int y = 0; y<m_size.y(); y++)
        {
            for (int x = 0; x<m_size.x(); x++)
            {
                putchar (GetValue(Vec2n(x,y)) ? 'X' : '.');
            }
            printf ("\n");
        }
    }
    Matrix<T> GetDeepCopy ()
    {
        Matrix<T> mat;
        mat.m_org = m_org;
        mat.m_size = m_size;
        mat.m_dirX = m_dirX;
        mat.m_dirY = m_dirY;
        mat.m_pdata = new MatrixData<T> (*this->m_pdata);
        return mat;
    }

    Matrix<T> GetShallowCopy ()
    {
        Matrix<T> mat = *this;
        return mat
    }

    // ShallowCopy
    Matrix& operator= (Matrix<T>& mat)
    {
        this->m_org = mat.m_org;
        this->m_size = mat.m_size;
        this->m_dirX = mat.m_dirX;
        this->m_dirY = mat.m_dirY;
        if (this->m_pdata)
            this->m_pdata->Unref();
        this->m_pdata = mat.m_pdata;
        if (this->m_pdata)
            this->m_pdata->Ref();
        return *this;
    }

    void GetSubMatrix (Matrix& out_mat, Vec2n org, Vec2n size)
    {
        //TODO
        Vec2n _org = m_org+(m_dirX*org.x()+m_dirY*org.y());
        if (_org.x() > m_pdata->GetSize().x() || _org.x() < 0)
            throw new ExceptionInvalidParam ("Param org.x is invalid");
        if (_org.y() > m_pdata->GetSize().y() || _org.y() < 0)
            throw new ExceptionInvalidParam ("Param org.y is invalid");
        Vec2n otherPos = _org + m_dirX*(size.x()-1) + m_dirY*(size.y()-1);
        if (otherPos.x() > m_pdata->GetSize().x() || otherPos.x() < 0)
            throw new ExceptionInvalidParam ("Param org.x/size.x is invalid");
        if (otherPos.y() > m_pdata->GetSize().y() || otherPos.y() < 0)
            throw new ExceptionInvalidParam ("Param org.y/size.y is invalid");

        if (out_mat.m_pdata)
            out_mat.m_pdata->Unref();
        out_mat.m_pdata = m_pdata;                
        out_mat.m_pdata->Ref();
        out_mat.m_size = size;
        out_mat.m_org = _org;
        out_mat.m_dirX = m_dirX;
        out_mat.m_dirY = m_dirY;
    }

    std::vector<int> GetNumSimilarRows (int minNumDifferences = 0, int maxNumDifferences = 0)
    {
        std::vector<int> vec;
        for (int y1=0,y2=1; y1 < m_size.y()-1; y1++, y2++)
        {
            int num = 0;
            for (int x=0; x < m_size.x(); x++)
            {
                if (GetValue (x, y1) != GetValue (x, y2))
                    num++;
                if (x+1 == m_size.x())
                    if ((num >= minNumDifferences) && (num <= maxNumDifferences))
                        vec.push_back (y2);
            }
        }
        return vec;
    }

    std::vector<int> GetNumSimilarColumns (int minNumDifferences = 0, int maxNumDifferences = 0)
    {
        // y and x are swapped
        std::vector<int> vec;
        for (int x1=0,x2=1; x1 < m_size.x()-1; x1++, x2++)
        {
            int num = 0;
            for (int y=0; y < m_size.y(); y++)
            {
                if (GetValue (x1, y) != GetValue (x2, y))
                    num++;
                if (y+1 == m_size.y())
                    if ((num >= minNumDifferences) && (num <= maxNumDifferences))
                        vec.push_back (x2);
            }
        }
        return vec;
    }


    std::vector<int> GetNumOneValueRows (type_t value, int minNumDifferences = 0, int maxNumDifferences = 0)
    {
        std::vector<int> vec;
        for (int y=0; y < m_size.y(); y++)
        {
            int num = 0;
            for (int x=0; x < m_size.x(); x++)
            {
                if (GetValue (x, y) != value)
                    num++;
                if (x+1 == m_size.x())
                    if ((num >= minNumDifferences) && (num <= maxNumDifferences))
                        vec.push_back (y);
            }
        }
        return vec;
    }

    std::vector<int> GetNumOneValueRowsIgnoreBorders (type_t value, int minNumDifferences = 0, int maxNumDifferences = 0)
    {
        std::vector<int> vec (GetNumOneValueRows (value, minNumDifferences, maxNumDifferences));
        std::list<int> list (vec.begin(), vec.end()); 
        for (int i = 0; !list.empty() && list.front() == i; i++)
            list.pop_front ();
        for (int i = m_size.y()-1; !list.empty() && (list.back() == i); i--)
            list.pop_back ();
        std::vector<int> retvec (list.begin(), list.end());
        return retvec;
    }

    std::vector<int> GetNumOneValueColumns (type_t value, int minNumDifferences = 0, int maxNumDifferences = 0)
    {
        std::vector<int> vec;
        for (int x=0; x < m_size.x(); x++)
        {
            int num = 0;
            for (int y=0; y < m_size.y(); y++)
            {
                if (GetValue (x, y) != value)
                    num++;
                if (y+1 == m_size.y())
                    if ((num >= minNumDifferences) && (num <= maxNumDifferences))
                        vec.push_back (x);
            }
        }
        return vec;
    }

    std::vector<int> GetNumOneValueColumnsIgnoreBorders (type_t value, int minNumDifferences = 0, int maxNumDifferences = 0)
    {
        std::vector<int> vec (GetNumOneValueRows (value, minNumDifferences, maxNumDifferences));
        std::list<int> list (vec.begin(), vec.end());        
        for (int i = 0; !list.empty() && list.front() == i; i++)
            list.pop_front ();
        for (int i = m_size.x()-1; !list.empty() && (list.back() == i); i--)
            list.pop_back ();
        std::vector<int> retvec (list.begin(), list.end());
        return retvec;
    }

    void LoadFromBMPFile (CSTR file, int step)
    {
        BMP bmp;
        BMPErr err = bmp.LoadFromFile (file);
        if (err == IMG_ERR_NO_FILE)
        {
            std::string msg = file;
            msg += " cannot be found";
            throw new ExceptionFileNotFound (msg.c_str());
        }
        else if (err == IMG_ERR_UNSUPPORTED || err == IMG_ERR_UNSUPPORTED)
        {
            throw new ExceptionUnsupported ("This type of BMP is not supported");
        }
        else // IMG_OK:
        {
            this->Recreate (bmp.GetWidth()/step, bmp.GetHeight()/step);
            for (int iny = step/2, outy=bmp.GetHeight()/step-1; iny < (signed int)bmp.GetHeight(); iny += step, outy--)
                for (int inx = step/2, outx=0; inx < (signed int)bmp.GetWidth(); inx += step, outx++)
                {
                    bool isBlack = bmp.IsPixelBlack (Vec2n (inx,iny));
                    this->SetValue (outx, outy, isBlack);
                }
        }
    }

    // return number of differences
    int CompareToMatrix (Matrix& mat)
    {
        int retval = 0;
        int countX = std::min (mat.GetSize().x(),this->GetSize().x());
        int countY = std::min (mat.GetSize().y(),this->GetSize().y());
        for (int x = 0; x < countX; x++)
            for (int y = 0; y < countY; y++)
            {
                type_t v1 = this->GetValue (x, y);
                type_t v2 = mat.GetValue (x, y);
                if (v1 != v2)
                    retval++;
            }
        return retval;
    }

    std::vector<Vec2n> FindPattern (Matrix& pattern, int acceptedDifferences = 0)
    {
        Vec2n patternSize = pattern.GetSize ();
        Vec2n mySize = this->GetSize ();
        if (patternSize.x() > mySize.x() || patternSize.y() > mySize.y())
            throw new ExceptionInvalidParam ("Size of pattern is bigger than size of matrix");
        std::vector<Vec2n> ret;

        for (int y = 0; y <= mySize.y()-patternSize.y(); y++)
            for (int x = 0; x <= mySize.x()-patternSize.x(); x++)
            {
                Matrix<bool> submat;
                this->GetSubMatrix (submat, Vec2n(x,y), patternSize);
                int differences = submat.CompareToMatrix (pattern);
                if (differences <= acceptedDifferences)
                    ret.push_back (Vec2n (x, y));
            }
        return ret;
    }
};