#pragma once


#include <string>
#include "qrcode.h"
#include "codewordbitstream.h"
#include "staticdata.h"
#include "factory.h"
#include "codeworddecoder.h"
#include "tableforalphanumMode.h"

class QRDataDecoder
{
    StaticData* m_sd;

public:
    QRDataDecoder ()
    {
        Factory factory;
        m_sd = factory.GetStaticData ();
    }
    typedef std::list<bool> buffer_t;
private:
    enum mode_e
    {
        ModeTerminator = 0,
        ModeNum = 1,
        ModeAlphanum = 2,
        ModeAppend = 3,
        Mode8bit = 4, 
        ModeFCN11 = 5,
        ModeECI = 7,   
        ModeKanji = 8, 
        ModeFCN12 = 9
    };
    void LoadDemaskedBits (std::vector<bool>& out_bits, QRCode& qrcode, int version, int mask, unsigned char errlvl)
    {
        const ErrorCharacteristic* errchar = m_sd->GetErrorCharacteristic (version, errlvl);

        int numCodewords = errchar->numTotalCodewords;
        int numBits = numCodewords * 8;
        out_bits.resize (numBits);

        CodewordBitstream bitstream (qrcode);
        for (int i = 0; i < numBits; i++)
        {
            bool value = bitstream.GetBit ();
            int posX = bitstream.GetPosX ();
            int posY = bitstream.GetPosY ();
            bool bmask = m_sd->GetMaskAtPos (posX, posY, mask);
            out_bits[i] = value ^ bmask;
            bitstream.NextBit ();
        }
    }
    int GetAndPopBits (buffer_t& bits, int countBits)
    {
        if (countBits <= 0 || countBits >= 32)
            throw new ExceptionInvalidParam ("Param countBits is invalid");
        
        int retval = 0;
        for (int i = 0; i < countBits; i++)
        {
            if (bits.empty ())
                throw new ::ExceptionOutOfBuffer ("Param bits is empty. It should not happened");
            retval = (retval<<1)|(int)bits.front ();
            bits.pop_front ();
        }
        return retval;
    }
    bool GetECIHeaderAndPopBits (buffer_t& bits, mode_e& out_mode)
    {
        int nmode = GetAndPopBits (bits, 4);
        switch (nmode)
        {
        case ModeTerminator:
        case ModeNum:
        case ModeAlphanum:
        case ModeAppend:
        case Mode8bit:
        case ModeFCN11:
        case ModeECI:   
        case ModeKanji:
        case ModeFCN12:
            out_mode = (mode_e) nmode;
            break;
        default:
            return false;
        }
        return true;
    }

    int  GetCharCountCountIndicatorSize (int version, mode_e mode)
    {
        if (version >= 1 && version <= 9)
        {
            switch (mode)
            {
            case ModeNum: return 10;
            case ModeAlphanum: return 9;
            case Mode8bit: return 8;
            case ModeKanji: return 8;
            }
        }
        else if (version >= 10 && version <= 26)
        {
            switch (mode)
            {
            case ModeNum: return 12;
            case ModeAlphanum: return 11;
            case Mode8bit: return 16;
            case ModeKanji: return 10;
            }
        }
        else if (version >= 27 && version <= 40)
        {
            switch (mode)
            {
            case ModeNum: return 14;
            case ModeAlphanum: return 13;
            case Mode8bit: return 16;
            case ModeKanji: return 12;
            }
        }
        throw new ExceptionInvalidParam ("Param version or mode is invalid");
    }

    int GetCharCountAndPopBits (buffer_t& bits, int version, mode_e mode)
    {
        int numBits = GetCharCountCountIndicatorSize (version, mode);
        int size = GetAndPopBits (bits, numBits);
        return size;
    }
   
    // mode is already popped
    void DecodeNumber (buffer_t& bits, int version, std::string& out_strData)
    {
        int numChars = GetCharCountAndPopBits (bits, version, ModeNum);
        while (numChars > 3)
        {
            int num = GetAndPopBits (bits, 10);
            char c4[4];
            sprintf (c4, "%03d", num);
            out_strData += c4;
            numChars -= 3;
        }
        if (numChars == 2)
        {
            int num = GetAndPopBits (bits, 7);
            char c3[3];
            sprintf (c3, "%02d", num);
            out_strData += c3;
            numChars -= 2;
        }
        else if (numChars == 1)
        {
            int num = GetAndPopBits (bits, 4);
            char c2[2];
            sprintf (c2, "%01d", num);
            out_strData += c2;
            numChars -= 1;
        }
    }
    void Decode8bit (buffer_t& bits, int version, std::string& out_strData)
    {
        int numChars = GetCharCountAndPopBits (bits, version, Mode8bit);
        for (int i = 0; i < numChars; i++)
        {
            unsigned char c = GetAndPopBits (bits, 8);
            out_strData += c;
        }
    }
    void DecodeAlphanum (buffer_t& bits, int version, std::string& out_strData)
    {
        TableForAlphanumMode table;
        int numChars = GetCharCountAndPopBits (bits, version, ModeAlphanum);
        for (int i = 0; i < numChars; i+=2)
        {
            int dchar = GetAndPopBits (bits, 11);
            char c3[3] = {0};
            c3[0] = table.Decode(dchar / 45);
            c3[1] = table.Decode(dchar % 45);
            out_strData += c3;
        }
        if (numChars%2)
        {
            char c2[2] = {0};
            c2[0] = table.Decode(GetAndPopBits (bits, 6));
            out_strData += c2;
        }
    }
public:
    bool Decode (QRCode& qrcode, unsigned char errlvl, int mask, std::string& out_strData)
    {
        int version = qrcode.GetVersionFromNumModules ();
        const ErrorCharacteristic* errchar = m_sd->GetErrorCharacteristic (version, errlvl);

        std::vector<bool> vecAllBits;
        LoadDemaskedBits (vecAllBits, qrcode, version, mask, errlvl);

        std::vector<bool> vecDataBits;
        CodewordDecoder decoder;
        if (! decoder.DecodeDataCodewords (vecAllBits, vecDataBits, version, errlvl))
            return false;

        buffer_t bufferBits (vecDataBits.begin(), vecDataBits.end());

        out_strData.clear ();
        
        while (true)
        {
            mode_e mode;
            if (! GetECIHeaderAndPopBits (bufferBits, mode))
                return false;
            switch (mode)
            {
            case ModeNum:
                DecodeNumber (bufferBits, version, out_strData);
                break;
            case Mode8bit:
                Decode8bit (bufferBits, version, out_strData);
                break;
            case ModeAlphanum:
                DecodeAlphanum (bufferBits, version, out_strData);
                break;
            case ModeTerminator:
                return true;
            default:
                throw new ExceptionUnsupported ("This mode is not supported");
                return false;
            }
        }

        return true;
    }
};