#include <stdlib.h>
#include <ostream>

#include <algorithm>
#include "factory.h"
#include "staticdata.h"
#include "qrcode.h"
#include "functionpatterns.h"
#include "qrdatadecoder.h"


const int c_sizeOfModul = 10;
const bool c_isBruteForceEnabled = true;

using std::cout; 
using std::endl;

void printVectorIfNotEmpty (CSTR msg, std::vector<int>& vec)
{
    if (! vec.empty ())
    {
        cout << msg << ": ";
        for (size_t i = 0; i < vec.size(); i++)
        {
            cout << vec[i] << " ";
        }
        cout << endl;
    }
}


bool CheckPositionPatterns (std::vector<Vec2n>& vecV2)
{
    Factory factory;
    StaticData* staticData = factory.GetStaticData ();

    if (vecV2.size() != 3)
        cout << "PROBLEM: Number of position patterns is invalid" << endl;
    if (vecV2.size() > 3)
        cout << "PROBLEM: Too many position patterns are found" << endl;
    if (vecV2.size () == 0)
        cout << "PROBLEM: No position pattern is found" << endl;
    if (vecV2.size () == 1)
        cout << "PROBLEM: Only one position pattern is found" << endl;
    bool validPos = true;
    if (vecV2.size () >= 1 && vecV2.size () <= 3)
    {
        Vec2n v[3];
        v[0] = v[1] = v[2] = vecV2[0];
        if (vecV2.size () >= 2)
            v[1] = vecV2[1];
        if (vecV2.size () >= 3)
            v[2] = vecV2[2];

        int maxDiff = 0;
        for (int i = 0; i < 3; i++)
            for (int j = i+1; j < 3; j++)
            {
                Vec2n diff = v[j] - v[i];
                diff.x() = diff.x() ? std::abs (diff.x())+7 : 0;
                diff.y() = diff.y() ? std::abs (diff.y())+7 : 0;
                maxDiff = std::max (diff.x(), diff.y());
            }
        for (int i = 0; i < 3; i++)
            for (int j = i+1; j < 3; j++)
            {
                Vec2n diff = v[j] - v[i];
                diff.x() = diff.x() ? std::abs (diff.x())+7 : 0;
                diff.y() = diff.y() ? std::abs (diff.y())+7 : 0;
                if (diff.x() != 0 && diff.x() != maxDiff && diff.y() != 0 && diff.y() != maxDiff)
                    validPos = false;
            }
        if (! staticData->IsCountModulesPerSideValid (maxDiff))
            validPos = false;
        if (! validPos)
            cout << "PROBLEM: Position patterns has invalid position" << endl;
    }
    return validPos && vecV2.size()==3;
}


void DecodeDataWithExceptionHandling (QRCode& qrcode, unsigned char errlvl, int mask, std::string& out_strData)
{
    bool berror = false;
    try
    {
        QRDataDecoder decoder;
        decoder.Decode (qrcode, errlvl, mask, out_strData);
    } 
    catch (ExceptionUnsupported*)
    {
        berror = true;
        cout << "PROBLEM: Unsupported mode in encoded data (?error?)" << endl;
    }
    catch (ExceptionOutOfBuffer*)
    {
        berror = true;
        cout << "PROBLEM: Encoded data are corrupted" << endl;
    }
    catch (...)
    {
        berror = true;
        cout << "PROBLEM: Something terrible happened during decoding data" << endl;
    }
    if (out_strData.empty ())
    {
        berror = true;
        cout << "PROBLEM: Length of data is zero ? Error!" << endl;
    }
    if (berror)
    {
        cout << "PROBLEM: Error has occured during decoding data" << endl;
    }
}

QRCode FindBestOrientation (QRCode& in)
{
    Matrix<bool> mat = in;
    int bestScore = in.GetNumInvalidPositionModulesWithSeparators ();
    int bestI = 0;

    for (int i = 0; i < 4; i++)
    {
        QRCode qrcode (mat);
        int score = qrcode.GetNumInvalidPositionModulesWithSeparators ();
        if (score < bestScore)
        {
            bestScore = score;
            bestI = i;
        }
        mat = mat.GetRotated ();
    }
    //mat = mat.GetVerticalFlippedMatrix ();
    
    for (int i = 0; i < bestI; i++)
        mat = mat.GetRotated ();

    return mat;
}

void PrintQRCodeStats (QRCode& qrcode)
{
    Factory factory;
    StaticData* pdata = factory.GetStaticData ();
    cout << "Total number of modules: " << qrcode.GetNumModulesPerSide() << "x" << qrcode.GetNumModulesPerSide() << endl;
    cout << "Version: " << qrcode.GetVersionFromNumModules () << endl;
    cout << "Invalid position patterns: " << qrcode.GetNumInvalidPositionModules () << 
                    "/" << qrcode.GetTotalNumPositionModules () << endl;
    cout << "Invalid position patterns with separators: " << qrcode.GetNumInvalidPositionModulesWithSeparators () << 
                    "/" << qrcode.GetTotalNumPositionModulesWithSeparators () << endl;
    cout << "Invalid alignment patterns: " << qrcode.GetNumInvalidAlignmentModules () << 
                    "/" << qrcode.GetTotalNumAlignmentModules () << endl;
    cout << "Invalid timing patterns: " << qrcode.GetNumInvalidTimingModules () <<
                    "/" << qrcode.GetTotalNumTimingModules () << endl;
    cout << "Invalid function patterns with separators: "
            <<  qrcode.GetNumInvalidPositionModulesWithSeparators () + qrcode.GetNumInvalidAlignmentModules () + qrcode.GetNumInvalidTimingModules ()
            << "/" <<  qrcode.GetTotalNumPositionModulesWithSeparators () + qrcode.GetTotalNumTimingModules () + qrcode.GetTotalNumAlignmentModules ()
            << endl;

    char errlvlLT = qrcode.GetErrorLevelLT();
    char errlvlRTLB = qrcode.GetErrorLevelRTLB();
    cout << "Error level LT: " << errlvlLT << endl;
    cout << "Error level RT/LB: " << errlvlRTLB << endl;
    if (errlvlLT != errlvlRTLB)
        cout << "PROBLEM: Error level LT != Error version RT/LB" << endl;

    int maskLT = qrcode.GetNumMaskPatternLT();
    int maskRTLB = qrcode.GetNumMaskPatternRTLB();
    cout << "Mask pattern LT: " << maskLT << endl;
    cout << "Mask pattern RT/LB: " << maskRTLB << endl;
    if (maskLT != maskRTLB)
        cout << "PROBLEM: Mask pattern LT != Mask pattern RT/LB" << endl;

    if (qrcode.IsVersionInfo ())
    {
        int verRT = qrcode.GetVersionFromVersionInfoRT();
        int verLB = qrcode.GetVersionFromVersionInfoRT();
        int verModules = qrcode.GetVersionFromNumModules();
        cout << "Version from version info RT: " << verRT << endl;
        cout << "Version from version info LB: " << verLB << endl;
        if (verRT != verLB)
            cout << "PROBLEM: Version info RT != Version info LB" << endl;
        if (verModules != verRT)
            cout << "PROBLEM: Version modules != Version info RT" << endl;
        if (verModules != verLB)
            cout << "PROBLEM: Version modules != Version info LB" << endl;
    }
    else
    {
        cout << "Version from version info RT: " << "NONE (Version<7)" << endl;
        cout << "Version from version info LB: " << "NONE (Version<7)" << endl;
    }

    cout << "Decoded data (maskLT='" << maskLT << "'; errlvlLT='" << errlvlLT << "')" << endl;
    std::string strData;
    DecodeDataWithExceptionHandling (qrcode, errlvlRTLB, maskRTLB, strData);

    cout << "DATA='" << strData << "'" << endl; 
    if (maskLT != maskRTLB || errlvlLT != errlvlRTLB)
    {
        cout << "PROBLEM: (Mask pattern LT != Mask pattern RT/LB) or (Error level LT != Error level RT/LB)" << endl; 
        cout << "Decoded data (maskRTLB='" << maskRTLB << "'; errlvlRTLB='" << errlvlRTLB << "')" << endl;
        std::string strData;
        DecodeDataWithExceptionHandling (qrcode, errlvlRTLB, maskRTLB, strData);
        cout << "DATA='" << strData << "'" << endl;     
    }
    cout << endl;

    //cout << "Codeword order:" << endl;
    //qrcode.PrintOrderOfCodewords ();
    //cout << endl;

    //cout << "Bit order:" << endl;
    //qrcode.PrintOrderOfBits ();
    //cout << endl;
}

void BruteForce (Matrix<bool>& mat, Vec2n& v2org, Vec2n& v2size)
{
    double best=10000000.0;
    int bestOrgX, bestOrgY, bestSizeX, bestSizeY;
    for (int size = 21; size <= 177; size+=4)
    {
        if (size > mat.GetSize().x() || size > mat.GetSize().y())
            break;

        int total;
        {
            Matrix<bool> tmpMat;
            mat.GetSubMatrix (tmpMat, Vec2n(0, 0), Vec2n(size, size));
            QRCode qrcode (tmpMat);
            total = qrcode.GetTotalNumPositionModulesWithSeparators () 
                        + qrcode.GetTotalNumTimingModules () 
                        + qrcode.GetTotalNumAlignmentModules ();
        }
        for (int x = 0; x <= mat.GetSize().x()-size; x++)
        {
            if (size+x > mat.GetSize().x())
                break;
            for (int y = 0; y <= mat.GetSize().y()-size; y++)
            {
                if (size+y > mat.GetSize().y())
                    break;
                Matrix<bool> tmpMat;
                mat.GetSubMatrix (tmpMat, Vec2n(x, y), Vec2n(size, size));
                QRCode qrcode (tmpMat);
                tmpMat = FindBestOrientation ((QRCode) tmpMat);
                int invalid = qrcode.GetNumInvalidPositionModulesWithSeparators () 
                            + qrcode.GetNumInvalidAlignmentModules () 
                            + qrcode.GetNumInvalidTimingModules ();
                if ((double)invalid/(double)total <= best)
                {
                    bestOrgX = x;
                    bestOrgY = y;
                    bestSizeX = size;
                    bestSizeY = size;
                    best = (double)invalid/(double)total;
                }
            }
        }
    }
    v2org = Vec2n (bestOrgX, bestOrgY);
    v2size = Vec2n (bestSizeX, bestSizeY);
}

void PrintRowsAndColumnsProblems (QRCode& qrcode)
{
    printVectorIfNotEmpty ("PROBLEM: Duplicit rows (this and prev)", qrcode.GetNumSimilarRows ());
    printVectorIfNotEmpty ("PROBLEM: Duplicit columns (this and prev)", qrcode.GetNumSimilarColumns ());
    printVectorIfNotEmpty ("PROBLEM: Similar rows (this and prev)(differences <= 4)", qrcode.GetNumSimilarRows (0, 4));
    printVectorIfNotEmpty ("PROBLEM: Similar columns (this and prev)(differences <= 4)", qrcode.GetNumSimilarColumns (0, 4));
    printVectorIfNotEmpty ("PROBLEM: White rows", qrcode.GetNumOneValueRows (false));
    printVectorIfNotEmpty ("PROBLEM: Black rows", qrcode.GetNumOneValueRows (true));
    printVectorIfNotEmpty ("PROBLEM: White columns", qrcode.GetNumOneValueColumns (false));
    printVectorIfNotEmpty ("PROBLEM: Black columns", qrcode.GetNumOneValueColumns (true));
    printVectorIfNotEmpty ("PROBLEM: White rows (differences <= 4)", qrcode.GetNumOneValueRows (false, 0, 4));
    printVectorIfNotEmpty ("PROBLEM: Black rows (differences <= 4)", qrcode.GetNumOneValueRows (true, 0, 4));
    printVectorIfNotEmpty ("PROBLEM: White columns (differences <= 4)", qrcode.GetNumOneValueColumns (false, 0, 4));
    printVectorIfNotEmpty ("PROBLEM: Black columns (differences <= 4)", qrcode.GetNumOneValueColumns (true, 0, 4));
    printVectorIfNotEmpty ("PROBLEM: White rows ignore borders", qrcode.GetNumOneValueRowsIgnoreBorders (false));
    printVectorIfNotEmpty ("PROBLEM: Black rows ignore borders", qrcode.GetNumOneValueRowsIgnoreBorders (true));
    printVectorIfNotEmpty ("PROBLEM: White columns ignore borders", qrcode.GetNumOneValueColumnsIgnoreBorders (false));
    printVectorIfNotEmpty ("PROBLEM: Black columns ignore borders", qrcode.GetNumOneValueColumnsIgnoreBorders (true));
    printVectorIfNotEmpty ("PROBLEM: White rows ignore borders (differences <= 4)", qrcode.GetNumOneValueRowsIgnoreBorders (false, 0, 4));
    printVectorIfNotEmpty ("PROBLEM: Black rows ignore borders (differences <= 4)", qrcode.GetNumOneValueRowsIgnoreBorders (true, 0, 4));
    printVectorIfNotEmpty ("PROBLEM: White columns ignore borders (differences <= 4)", qrcode.GetNumOneValueColumnsIgnoreBorders (false, 0, 4));
    printVectorIfNotEmpty ("PROBLEM: Black columns ignore borders (differences <= 4)", qrcode.GetNumOneValueColumnsIgnoreBorders (true, 0, 4));
}

int main (int argc, const char** argv)
{
#ifndef DEBUG
    try 
    {
#endif
    InitLib (".");
    if (argc == 1)
        throw new ExceptionInvalidParam ("Program argument is missing");

    Factory factory;
    StaticData* staticData = factory.GetStaticData ();
 
    QRCode qrcode;

    cout << "Loading file '" << argv[1] << "'" << endl;
    qrcode.LoadFromBMPFile (argv[1], c_sizeOfModul);

    PrintRowsAndColumnsProblems (qrcode);

    FunctionPatterns fp;
    std::vector<Vec2n> vecV2 = qrcode.FindPattern (fp.GetPositionPattern (), 4);
    cout << "Position patterns (differces <= 4): ";
    for (size_t i = 0; i < vecV2.size(); i++)
        cout << vecV2[i].x()-1 << "x" << vecV2[i].y()-1 << " ";
    cout << endl;

    bool isAlignOK;
    Vec2n v2min;
    Vec2n v2max;
    if (isAlignOK = CheckPositionPatterns (vecV2))
    {
        cout << endl << "Position patterns are valid. Let's go decode..." << endl;
        v2min = vecV2[0];
        v2max = vecV2[0];
        for (int i = 1; i < 3; i++)
        {
            v2min = Vec2n (std::min (v2min.x(), vecV2[i].x()), std::min (v2min.y(), vecV2[i].y()));
            v2max = Vec2n (std::max (v2max.x(), vecV2[i].x()), std::max (v2max.y(), vecV2[i].y()));
        }
        v2max.x() += 7;
        v2max.y() += 7;
        cout << "Found QRcode from " << v2min.x() << "x" << v2min.y() << " to " << v2max.x() << "x" << v2max.y() << endl;
        
        Matrix<bool> submat;
        qrcode.GetSubMatrix (submat, v2min, v2max-v2min);
        submat = FindBestOrientation ((QRCode) submat);
        PrintQRCodeStats (QRCode (submat));
    }

    if (c_isBruteForceEnabled)
    {
        cout << "Brute force? Let's try..." << endl;
        Vec2n v2BForg;
        Vec2n v2BFsize;
        BruteForce (qrcode, v2BForg, v2BFsize);
        Vec2n v2BFmin = v2BForg;
        Vec2n v2BFmax = v2BForg + v2BFsize;
        cout << "Found QRcode from " << v2BFmin.x() << "x" << v2BFmin.y() << " to " << v2BFmax.x() << "x" << v2BFmax.y() << endl;

        if (isAlignOK && (v2min != v2BFmin || v2max != v2BFmax))
            cout << "PROBLEM: Result of brute force != Result of recommended sequence" << endl;

        Matrix<bool> submat;
        qrcode.GetSubMatrix (submat, v2BForg, v2BFsize);
        qrcode = FindBestOrientation (qrcode);
        PrintQRCodeStats (QRCode (submat));
    }
#ifndef DEBUG
    } 
    catch (Exception* ex)
    {
        cout << "ERROR: " << ex->what() << endl;
    }
    catch (...)
    {
        cout << "ERROR: Error is not resolved" << endl;
    }
#endif
    DestroyLib ();

    return EXIT_SUCCESS;
}

